"""
This module contains utility functions to work with Gitmoji objects.
"""

from typing import List

from shared.model import Gitmoji
from shared.spec import mojis

# global pattern to validate commit messages
PATTERN = (
    # To explicitly make . match new line
    r"(?s)"
    #  type
    r"({type})"
    # scope
    r"(\(\S+\))?!?:"
    # gitmoji and subject
    r"( {icon}(?=\s)| [^\n\r]+)"
    # body
    r"((\n\n.*)|(\s*))?$"
)


def get_gitmojis() -> List[Gitmoji]:
    """Return the list of Gitmoji objects."""
    return [Gitmoji(**moji) for moji in mojis]


def get_type_pattern() -> str:
    """Return the type patterns."""
    return "|".join([moji.type for moji in get_gitmojis()])


def get_icon_pattern() -> str:
    """Return icon patterns."""
    return "|".join([moji.icon for moji in get_gitmojis()])


def get_pattern() -> str:
    """Return the complete validation pattern."""
    type_pattern = get_type_pattern()
    icon_pattern = get_icon_pattern()

    return PATTERN.replace("type", type_pattern).replace("icon", icon_pattern)



"""
This module contains the Gitmoji class.
"""

from shared.gitmojis import GitMojiConstant


mojis = [
    {
        "type": "fix",
        "icon": GitMojiConstant.GJ_FIX.value,
        "code": ":bug:",
        "desc": "Fix a bug.",
    },
    {
        "type": "feat",
        "icon": GitMojiConstant.GJ_FEAT.value,
        "code": ":sparkles:",
        "desc": "Introduce new features.",
    },
    {
        "type": "docs",
        "icon": GitMojiConstant.GJ_DOCS.value,
        "code": ":memo:",
        "desc": "Add or update documentation.",
    },
    {
        "type": "style",
        "icon": GitMojiConstant.GJ_STYLE.value,
        "code": ":art:",
        "desc": "Improve structure / format of the code.",
    },
    {
        "type": "refactor",
        "icon": GitMojiConstant.GJ_REFACTOR.value,
        "code": ":recycle:",
        "desc": "Refactor code.",
    },
    {
        "type": "perf",
        "icon": GitMojiConstant.GJ_PERF.value,
        "code": ":zap:",
        "desc": "Improve performance.",
    },
    {
        "type": "test",
        "icon": GitMojiConstant.GJ_TEST.value,
        "code": ":white_check_mark:",
        "desc": "Add, update, or pass tests.",
    },
    {
        "type": "build",
        "icon": GitMojiConstant.GJ_BUILD.value,
        "code": ":construction_worker:",
        "desc": "Add or update CI build system.",
    },
    {
        "type": "ci",
        "icon": GitMojiConstant.GJ_CI.value,
        "code": ":green_heart:",
        "desc": "Fix CI Build.",
    },
    {
        "type": "dump",
        "icon": GitMojiConstant.GJ_DUMP.value,
        "code": ":fire:",
        "desc": "Remove code or files.",
    },
    {
        "type": "hotfix",
        "icon": GitMojiConstant.GJ_HOTFIX.value,
        "code": ":ambulance:",
        "desc": "Critical hotfix.",
    },
    {
        "type": "init",
        "icon": GitMojiConstant.GJ_INIT.value,
        "code": ":tada:",
        "desc": "Begin a project.",
    },
    {
        "type": "bump",
        "icon": GitMojiConstant.GJ_BUMP.value,
        "code": ":bookmark:",
        "desc": "Release / Version tags.",
    },
    {
        "type": "dep-drop",
        "icon": GitMojiConstant.GJ_DEP_DROP.value,
        "code": ":arrow_down:",
        "desc": "Downgrade dependencies.",
    },
    {
        "type": "dep-bump",
        "icon": GitMojiConstant.GJ_DEP_BUMP.value,
        "code": ":arrow_up:",
        "desc": "Upgrade dependencies.",
    },
    {
        "type": "dep-add",
        "icon": GitMojiConstant.GJ_DEP_ADD.value,
        "code": ":heavy_plus_sign:",
        "desc": "Add a dependency.",
    },
    {
        "type": "dep-rm",
        "icon": GitMojiConstant.GJ_DEP_RM.value,
        "code": ":heavy_minus_sign:",
        "desc": "Remove a dependency.",
    },
    {
        "type": "config",
        "icon": GitMojiConstant.GJ_CONFIG.value,
        "code": ":wrench:",
        "desc": "Add or update configuration files.",
    },
    {
        "type": "lang",
        "icon": GitMojiConstant.GJ_LANG.value,
        "code": ":globe_with_meridians:",
        "desc": "Internationalization and localization.",
    },
    {
        "type": "merge",
        "icon": GitMojiConstant.GJ_MERGE.value,
        "code": ":twisted_rightwards_arrows:",
        "desc": "Merge branches.",
    },
    {
        "type": "logs-add",
        "icon": GitMojiConstant.GJ_LOGS_ADD.value,
        "code": ":loud_sound:",
        "desc": "Add or update logs.",
    },
    {
        "type": "logs-rm",
        "icon": GitMojiConstant.GJ_LOGS_RM.value,
        "code": ":mute:",
        "desc": "Remove logs.",
    },
    {
        "type": "ignore",
        "icon": GitMojiConstant.GJ_IGNORE.value,
        "code": ":see_no_evil:",
        "desc": "Add or update a .gitignore file.",
    },
    {
        "type": "chore",
        "icon": GitMojiConstant.GJ_CHORE.value,
        "code": ":broom:",
        "desc": "Commit changes related to miscellaneous chores.",
    },
]



"""
This module contains the Gitmoji constants.
"""

from enum import Enum

__all__ = ["GitMojiConstant"]


class GitMojiConstant(Enum):
    """Enum class for Gitmojis."""

    GJ_FIX = "🐛"
    GJ_FEAT = "✨ "
    GJ_DOCS = "📝"
    GJ_STYLE = "🎨"
    GJ_REFACTOR = "♻️"
    GJ_PERF = "⚡️"
    GJ_TEST = "✅ "
    GJ_BUILD = "👷"
    GJ_CI = "💚"
    GJ_DUMP = "🔥"
    GJ_HOTFIX = "🚑️"
    GJ_INIT = "🎉"
    GJ_BUMP = "🔖"
    GJ_DEP_DROP = "⬇️"
    GJ_DEP_BUMP = "⬆️"
    GJ_DEP_ADD = "➕ "
    GJ_DEP_RM = "➖ "
    GJ_CONFIG = "🔧"
    GJ_LANG = "🌐"
    GJ_MERGE = "🔀"
    GJ_LOGS_ADD = "🔊"
    GJ_LOGS_RM = "🔇"
    GJ_IGNORE = "🙈"
    GJ_CHORE = "🧹"



"""
This module contains the CommitizenGitmojiCz class,
which is a subclass of BaseCommitizen.
"""

import re
import textwrap
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List

from commitizen.cz.base import BaseCommitizen
from commitizen.cz.utils import required_validator, multiple_line_breaker
from commitizen.defaults import MAJOR, MINOR, PATCH

from shared import utils
from shared.gitmojis import GitMojiConstant as mojis

__all__ = ["CommitizenGitmojiCz", "parse_scope", "parse_subject"]


def parse_scope(text: str) -> str:
    """Parse the scope of the change."""
    if not text:
        return ""

    scope = text.strip().split()
    return "-".join(scope)


def parse_subject(text: str) -> str:
    """Parse the subject of the change."""
    if isinstance(text, str):
        text = text.strip(".").strip()

    return required_validator(text, msg="Subject is required.")


def parse_bordy(text: str) -> str:
    """Process the text using multiple_line_breaker and then soft-wrap it."""
    processed_text = multiple_line_breaker(text)

    return "\n".join(textwrap.wrap(processed_text, width=72))


class CommitizenGitmojiCz(BaseCommitizen):
    """Commitizen adapter for gitmoji style."""

    # pattern to match messages for bumping
    # if none of these match, version will not be bumped
    # (unless manually specified)
    bump_pattern = (
        r"^((BREAKING[\-\ ]CHANGE"
        r"|feat"
        r"|fix"
        r"|hotfix"
        r"|refactor"
        r"|perf)"
        r"(\(.+\))?"  # scope
        r"!?):"  # breaking
    )
    # map types to SemVer
    bump_map = OrderedDict(
        (
            (r"^.+!$", MAJOR),
            (r"^BREAKING[\-\ ]CHANGE", MAJOR),
            (r"^feat", MINOR),
            (r"^fix", PATCH),
            (r"^hotfix", PATCH),
            (r"^refactor", PATCH),
            (r"^perf", PATCH),
        )
    )
    bump_map_major_version_zero = bump_map.copy()

    # parse information for generating the change log
    commit_parser = (
        rf"^(?P<change_type>{utils.get_type_pattern()}|BREAKING CHANGE)"
        rf"(?:\((?P<scope>[^()\r\n]*)\)|\()?(?P<breaking>!)?:\s"
        rf"(?P<message>{utils.get_icon_pattern()}\s.*)?"
    )
    # exclude from changelog
    changelog_pattern = r"^(?!init)(?!merge)(?!bump).*"
    # map types to changelog sections
    change_type_map = {
        # features
        "feat": f"{mojis.GJ_FEAT.value} Features",
        # fixes
        "fix": f"{mojis.GJ_FIX.value} Fixes",
        "hotfix": f"{mojis.GJ_HOTFIX.value} Fixes",
        # refactorings
        "refactor": f"{mojis.GJ_REFACTOR.value} Refactorings",
        # style & architecture
        "style": f"{mojis.GJ_STYLE.value} Style & Architecture",
        # performance
        "perf": f"{mojis.GJ_PERF.value} Performance",
        # docs
        "docs": f"{mojis.GJ_DOCS.value} Documentation",
        # tests
        "test": f"{mojis.GJ_TEST.value} Tests",
        # ci & build
        "build": f"{mojis.GJ_BUILD.value} Build",
        "ci": f"{mojis.GJ_CI.value}{mojis.GJ_BUILD.value} CI",
        # configuration & scripts & packages
        "config": f"{mojis.GJ_CONFIG.value} Configuration, Scripts, Packages",
        # cleanup
        "dump": f"{mojis.GJ_DUMP.value} Clean up",
        # dependencies
        "dep-add": f"{mojis.GJ_DEP_ADD.value} Dependencies",
        "dep-rm": f"{mojis.GJ_DEP_RM.value} Dependencies",
        "dep-bump": f"{mojis.GJ_DEP_BUMP.value} Dependencies",
        "dep-drop": f"{mojis.GJ_DEP_DROP.value} Dependencies",
        # language & accessibility
        "lang": f"{mojis.GJ_LANG.value} Language & Accessibility",
        # logs
        "logs-add": f"{mojis.GJ_LOGS_ADD.value} Logs",
        "logs-rm": f"{mojis.GJ_LOGS_RM.value} Logs",
        # ignore
        "ignore": f"{mojis.GJ_IGNORE.value} Ignore",
        # None: init, bump, merge
    }
    # Order sections in changelog
    # All other sections are ordered alphabetically
    change_type_order = [
        f"{mojis.GJ_FEAT.value} Features",
        f"{mojis.GJ_FIX.value}{mojis.GJ_HOTFIX.value} Fixes",
        f"{mojis.GJ_REFACTOR.value} Refactorings",
        f"{mojis.GJ_PERF.value} Performance",
    ]

    def questions(self) -> List[Dict[str, Any]]:
        """Return the questions to ask the user."""
        return [
            {
                "type": "list",
                "name": "prefix",
                "message": "Select the type of change you are committing",
                "choices": [
                    {
                        "value": moji.value,
                        "name": moji.name,
                    }
                    for moji in utils.get_gitmojis()
                ],
            },
            {
                "type": "input",
                "name": "scope",
                "message": (
                    "What is the scope of this change? (class or file name): "
                    "(press [enter] to skip)\n"
                ),
                "filter": parse_scope,
            },
            {
                "type": "input",
                "name": "subject",
                "filter": parse_subject,
                "message": (
                    "Write a short and imperative summary "
                    "of the code changes: "
                    "(lower case (except for name) and no period)\n"
                ),
            },
            {
                "type": "input",
                "name": "body",
                "message": (
                    "Provide additional contextual information about "
                    "the code changes: (press [enter] to skip)\n"
                ),
                "filter": parse_bordy,
            },
            {
                "type": "confirm",
                "message": (
                    "Is this a BREAKING CHANGE? "
                    "Correlates with MAJOR in SemVer"
                ),
                "name": "is_breaking_change",
                "default": False,
            },
            {
                "type": "input",
                "name": "footer",
                "message": (
                    "Footer. Information about Breaking Changes and reference "
                    "issues that this commit closes: (press [enter] to skip)\n"
                ),
            },
        ]

    def message(self, answers: Dict[str, Any]) -> str:
        """Generate a commit message from the answers."""
        prefix: tuple[str, str] = answers["prefix"]
        scope = answers["scope"]
        subject = answers["subject"]
        body = answers["body"]
        footer = answers["footer"]
        is_breaking_change = answers["is_breaking_change"]

        if scope:
            scope = f"({scope})"
        if body:
            body = f"\n\n{body}"
        if (
            is_breaking_change
            and footer
            and not footer.startswith("BREAKING CHANGE")
        ):
            footer = f"BREAKING CHANGE: {footer}"
        if footer:
            footer = f"\n\n{footer}"

        prefix_scope = f"{prefix[0]}{scope}"
        if is_breaking_change:
            prefix_scope = f"{prefix_scope}!"

        message = f"{prefix_scope}: {prefix[1]} {subject}{body}{footer}"

        return message

    def example(self) -> str:
        """Return an example commit message."""
        return (
            f"fix: {mojis.GJ_FIX.value} correct minor typos in code\n"
            "\n"
            "see the issue for details on the typos fixed\n"
            "\n"
            "closes issue #12"
        )

    def schema(self) -> str:
        """Return the commit message schema."""
        return (
            "<type>(<scope>): <gitmoji> <subject>\n"
            "<BLANK LINE>\n"
            "<body>\n"
            "<BLANK LINE>\n"
            "(BREAKING CHANGE: )<footer>"
        )

    # pattern to validate commits
    def schema_pattern(self) -> str:
        """Return the schema validation pattern."""
        return utils.get_pattern()

    def info(self) -> str:
        """Return information about the commit message style."""
        dir_path = Path(__file__).parent
        filepath = dir_path.joinpath("conventional_gitmojis_info.txt")
        with open(filepath, "r", encoding="UTF-8") as f:
            content = f.read()
        return content

    def process_commit(self, commit: str) -> str:
        """Process a commit."""
        pat = self.schema_pattern()
        m = re.match(pat, commit)
        if m is None:
            return ""
        return m.group(3).strip()
